import pandas as pd
import numpy as np

from .helpers import diag

def movement_splice(revisions: pd.DataFrame, window: int = 12) -> pd.DataFrame:
    """Apply the movement splice revision extension method.

    Parameters
    ----------
    revisions : pandas DataFrame
        A DataFrame of the rolling revisions. The values only need to be
        correct for the size of the moving window, as only the main
        diagonal and values shifted <= window length are used.
    window : int, default 12
        The size of the moving window excluding the base period.

    Returns
    -------
    pandas DataFrame
        The spliced index revisions. 
    """
    # Get movement splice by dividing by the previous revision value and
    # take the cumulative product.
    movement_splice = (
        revisions.div(revisions.shift(1))
        .pipe(diag)
        # The revisions begin after the first full window.
        .iloc[window+1:]
        .cumprod()
    )

    # Multiply by the value at the end of the initial revised window.
    return movement_splice.mul(revisions.iat[window, window])

def hasp(
    revisions: pd.DataFrame,
    window: int = 12,
) -> pd.DataFrame:
    """Apply the half window splice revision extension method.
    
    Parameters
    ----------
    revisions : pandas DataFrame
        A DataFrame of the rolling revisions. The values only need to be
        correct for the size of the moving window, as only the main
        diagonal and values shifted <= window length are used.
    window : int, default 12
        The size of the moving window excluding the base period.

    Returns
    -------
    pandas DataFrame
        The spliced index revisions. 
    """
    return wisp(revisions, window, half_window=True)

def wisp(
    revisions: pd.DataFrame,
    window: int = 12,
    half_window: bool = False,
) -> pd.DataFrame:
    """Apply the window splice revision extension method.

    Pass ``half_window=True`` for the half window splice.

    Parameters
    ----------
    %(splice.parameters)s
    half_window : bool, default False
        Computes half window splice when True.

    Returns
    -------
    %(splice.returns)s
    """
    # Need to retain original window len if half window is selected.
    full_window = window
    if half_window:
        window = int(np.floor(window/2))

    # Divide each rolling revision value by the base value.
    window_splice = (
        revisions.div(revisions.shift(window))
        .pipe(diag)
        # The revisions begin after the first full window.
        .iloc[full_window+1:]
        # Take cumprod every n periods where n is window length.
        .pipe(cumprod_over_periods, periods=window)
    )

    # Take full_window+1 as the initial window to include base period.
    initial_window = revisions.iloc[:full_window+1, full_window]

    # Repeat initial window values for the length of the revisions.
    repeated_window = repeat_values(
        # Using -window covers full and half window options.
        initial_window.iloc[-window:],
        # We only want to repeat it from the start of the revisions.
        index=window_splice.index,
    )

    # Multiply by repeated initial window to get revisions.
    return window_splice.mul(repeated_window)

def mean_pub(revisions: pd.DataFrame, window: int = 12) -> pd.DataFrame:
    """Apply the mean splice revision extension method.

    Parameters
    ----------
    revisions : pandas DataFrame
        A DataFrame of the rolling revisions. The values only need to be
        correct for the size of the moving window, as only the main
        diagonal and values shifted <= window length are used.
    window : int, default 12
        The size of the moving window excluding the base period.

    Returns
    -------
    pandas DataFrame
        The spliced index revisions. 
    """
    # Since the geometric mean is a product of all values raised to the
    # power of the reciprocal (of the window length), first multiply
    # all the terms together.

    # When considering a full window plus the base period value i.e. n+1
    # values, where n is the window length, take a rolling product of
    # the first n values which includes the base period and divide that
    # by the latest available revised value for that period i.e. the
    # (n+1)th value, raised to the power of n.
    rolling_splice = (
        revisions.pow(window)
        .div(
            revisions.rolling(window).apply(np.prod, raw=True)
            .shift(1)
        )
        .pipe(diag)
    )

    # Remove the datetime name of the initial window and convert to
    # frame so it can be multiplied by the rolling_splice.
    initial_window = (
        revisions.iloc[:window+1, window]
        .rename(None)
        .to_frame()
    )

    # Initialise the final indices with the initial_window.
    indices = initial_window.reindex(revisions.index)

    # The latest revised index value needs to be known for the following
    # steps. No simple vectorised method was found, so loop through
    # appending the latest value each time.
    for _ in range(len(revisions.index) - window):
        # Fill back with the initial window since rolling_splice only
        # has values in the revisions periods, which results in NAs in
        # the initial window after the final index calculation step.
        indices = indices.fillna(initial_window)

        # Get the additional multiplication terms in a similar way, by
        # taking a rolling product of the indices and dividing by the
        # latest revised value to the power of n, where n is the window
        # length.
        index_splice = (
            indices.rolling(window).apply(np.prod, raw=True)
            .div(indices.pow(window))
            # Shift to align with the latest period being calculated.
            .shift(1)
        )

        # Calculate the indices by multiplying all the values together
        # and raising to the power of the reciprocal of the window
        # length (1/n) to get the geometric mean, and multiplying the
        # result by the previous index value.
        indices = (
            rolling_splice.mul(index_splice).pow(1/window)
            .mul(indices.shift(1))
        )

    # Return only the revisions.
    return indices.iloc[window+1:]


def cumprod_over_periods(df: pd.DataFrame, periods: int = 12) -> pd.DataFrame:
    """Take the cumprod of values that share the same window period."""
    # Bin the values up by period number.
    periods = np.resize(range(periods), len(df))
    period_aligned = df.set_index(periods, append=True).unstack()
    output = period_aligned.cumprod()
    # Return to original shape and type.
    return output.stack().droplevel(-1).astype(df.dtypes)


def repeat_values(
    values,
    index,
) -> pd.DataFrame:
    """Repeat values to length of given index returning as DataFrame."""
    return pd.DataFrame(np.resize(values, len(index)), index=index)


from typing import List, Tuple

import pandas as pd
import numpy as np

__author__ = ['Dr. Usman Kayani']

def diag(
    df: pd.DataFrame
) -> pd.DataFrame:
    """
    Convert df to diagonal matrix.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to convert to diagonal matrix.

    Returns
    ------- 
    pd.DataFrame
        Diagonal matrix.
    """
    return pd.DataFrame(np.diag(df), index=df.index)

def _weights_calc(
    df: pd.DataFrame, 
    price_col: str='price',
    quantity_col: str='quantity',
    date_col: str='month',
    product_id_col: str='id',
) -> List:
    """
    Calculate weights from expenditure shares.
    
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to calculate weights from.
    price_col : str, optional
        Column name of price. The default is 'price'.
    quantity_col : str, optional
        Column name of quantity. The default is 'quantity'.
    date_col : str, optional
        Column name of date. The default is 'month'.
    product_id_col : str, optional
        Column name of product id. The default is 'id'.

    Returns
    ------- 
    List
        List of weights.
    """
    # Pivot dataframe for weights calculation.
    df_pivot = df.pivot(index=product_id_col, columns=date_col).fillna(0)

    # Set up expenditure and weights df from expenditure shares.
    expenditure = df_pivot[price_col] * df_pivot[quantity_col]
    weights = expenditure.divide(expenditure.sum())
    
    # Melt df for outpout.
    weights = pd.melt(
        weights.reset_index(),
        id_vars=[product_id_col],
        value_vars=df[date_col].unique(),
        var_name=date_col,
        value_name='weights',
    )

    # Merge back with input df.
    return pd.merge(
            right=weights, 
            left=df, 
            on=[date_col, product_id_col]
    )

def _vars_split(df) -> Tuple[List[str]]:
    """
    Split vars into categorical and numerical.
    
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to split.

    Returns
    -------
    Tuple[List[str]]
        Tuple of categorical and numerical vars.
    """
    numerical_vars = (
        df
        .select_dtypes(include='number')
        .columns
        .tolist()
    )
    categorical_vars = list(set(df.columns) - set(numerical_vars))   
    return categorical_vars, numerical_vars



"""
Provides the following functions:

* :func:`bilateral_methods`
* :func:`multilateral_methods`

With the following bilateral price index methods:

* Carli, Jevons, Dutot, Laspeyres, Paasche, geom_Laspeyres, geom_Paasche,
  Drobish, Marshall Edgeworth, Palgrave, Fisher, Tornqvist, Walsh, Sato Vartia,
  Geary Khamis, Rothwell.

and the following multilateral price index methods:

* The GEKS method paired with a bilateral method, the Geary-Khamis method (GK)
  and Time Dummy methods (TPD, TDH).
"""
from typing import Sequence, Optional, Union

import pandas as pd
import seaborn as sns

from .helpers import _weights_calc
from .bilateral import *
from .multilateral import *

__author__ = ['Dr. Usman Kayani']

def bilateral_methods(
    df: pd.DataFrame,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str='month',
    product_id_col: str='id',
    groups: Optional[Sequence[str]] = None,
    method: str = 'tornqvist',
    base_month: Union[int, str] = 1,
    reference_month: Union[int, str] = 1,
    plot: bool = False,
) -> pd.DataFrame:
    """
    Calculate all the bilateral indices.

    Parameters
    ----------
    df: pandas DataFrame
        Contains price and quantity columns, a time series column, and a product
        ID column as a minimum. A characteristics column should also be present
        for hedonic methods.
    price_col: str, defaults to 'price'
        User-defined name for the price column.
    quantity_col: str, defaults to 'quantity'
        User-defined name for the quantity column.
    date_col: str, defaults to 'month'
        User-defined name for the date column.
    product_id_col: str, defaults to 'id'
        User-defined name for the product ID column.
    groups: list of str, defaults to None
        The names of the groups columns.
    method: str, defaults to 'tornqvist'
        Options: {'carli', 'jevons', 'dutot', 'laspeyres', 'paasche',
        'geom_laspeyres', 'geom_paasche', 'drobish', 'marshall_edgeworth',
        'palgrave', 'fisher', 'tornqvist', 'walsh', 'sato_vartia', 'lowe',
        'geary_khamis_b', 'tpd', 'rothwell'}

        The bilateral method to use.
    base_month: int or str, defaults to 1
        Integer or string specifying the base month. An integer specifies the
        position while a string species the month in the format 'YYYY-MM'.
    reference_month: int or str, defaults to 1
        Integer or string specifying the reference month for rebasing if
        different from the base month. An integer specifies the position while a
        string species the month in the format 'YYYY-MM'.
    plot: bool, defaults to False
        Boolean parameter on whether to plot the resulting timeseries for price
        indices.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the timeseries and index values.
    """
    method = method.lower()

    valid_bilateral_methods = {
        'carli', 'jevons', 'dutot', 'laspeyres', 'lowe',
        'paasche', 'geom_laspeyres', 'geom_paasche', 'drobish',
        'marshall_edgeworth', 'palgrave', 'fisher', 'tornqvist',
        'walsh', 'sato_vartia', 'geary_khamis_b', 'tpd', 'rothwell'
    }

    if method not in valid_bilateral_methods:
        raise ValueError("Invalid option, please select a valid bilateral method.")

    args = (price_col, quantity_col, date_col, product_id_col)

    periods = sorted(df[date_col].unique())
    no_of_periods = len(periods)

    # Obtain the base period in the dataframe.
    base_period = periods[base_month-1]

    # Determine product IDs present in the base period.
    df_base = df.loc[df[date_col] == base_period]
    keep_ids = df_base.loc[:, product_id_col].unique()

    # Filter df to remove product IDs not present in the base period.
    df = df[df[product_id_col].isin(keep_ids)].reset_index(drop=True)

    if groups:
        return (
            df
            .groupby(groups)
            .apply(
                    lambda df_group: bilateral_methods(
                        df_group,
                        *args,
                        method=method,
                        base_month=base_month,
                        reference_month=reference_month,
                        plot=plot,
                    )
            )
            .reset_index()
            .rename({'level_1': 'month'}, axis=1)
        )

    if isinstance(base_month, str):
        base_month = periods.index(base_month) + 1

    if isinstance(reference_month, str):
        reference_month = periods.index(reference_month) + 1

    index_vals = np.zeros(no_of_periods)

    if method != 'tpd':
        # Obtain bilateral function for bilateral method.
        func = globals()[method]

    for i in range(no_of_periods):
        df_curr = df.loc[df[date_col] == periods[i]]

        # Make sure the sample is matched for given periods.
        df_base = df_base[df_base[product_id_col].isin(df_curr[product_id_col])]
        df_curr = df_curr[df_curr[product_id_col].isin(df_base[product_id_col])]

        if method == 'tpd':
            # Use multilateral TPD method with two periods.
            df_matched = (
                pd.concat([df_base, df_curr])
                .drop_duplicates()
            )
            # Recalculate weights for matched df.
            df_matched = _weights_calc(df_matched)
            # Append values to upper triangular of matrix.
            index_vals[i] = time_dummy(df_matched)[-1]
        else:
            # Find price and quantity vectors of base period and current period.
            p_base = df_base[price_col].to_numpy()
            p_curr = df_curr[price_col].to_numpy()
            data = (p_base, p_curr)

            # Get quantities for bilateral methods that use this information.
            if method in {
                'laspeyres', 'drobish', 'marshall_edgeworth',
                'geom_laspeyres', 'tornqvist', 'fisher',
                'walsh', 'sato_vartia', 'geary_khamis_b', 
                'rothwell', 'lowe'
            }:
                q_base = df_base[quantity_col].to_numpy()
                data += (q_base, )
            if method in {
                'paasche', 'drobish','palgrave',
                'marshall_edgeworth', 'geom_paasche', 'tornqvist',
                'fisher', 'walsh', 'sato_vartia',
                'geary_khamis_b'
            }:
                q_curr = df_curr[quantity_col].to_numpy()
                data += (q_curr, )

            # Determine the bilaterals for each base and current period and
            # append to upper tringular of matrix.
            index_vals[i] = func(*data)

    output_df = (
        pd.DataFrame(
            index_vals,
            index=periods
        )
        .rename({0: 'index_value'}, axis=1)
    )
    output_df.sort_index(inplace=True)
    if base_month != reference_month:
        # Rebase the index values to the reference month.
        output_df = output_df / output_df.iloc[reference_month-1]
    if plot:
        sns.set(rc={'figure.figsize':(11, 4)})
        (output_df * 100).plot(linewidth=2)
    return output_df

def multilateral_methods(
    df: pd.DataFrame,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str='month',
    product_id_col: str='id',
    characteristics: Optional[Sequence[str]] = None,
    groups: Optional[Sequence[str]] = None,
    method: str = 'all',
    bilateral_method: str = 'tornqvist',
    td_engine: str = 'numpy',
    refence_month: Union[int, str] = 1,
    plot: bool = False,
) -> pd.DataFrame:
    """
    Calculate all the multilateral indices.

    Currently supported: GEKS (geks), Geary-Khamis (gk), Time Product Dummy
    (tpd) and Time Hedonic Dummy (tdh).

    Parameters
    ----------
    df: pandas DataFrame
        Contains price and quantity columns, a time series column, and a product
        ID column as a minimum. A characteristics column should also be present
        for hedonic methods.
    price_col: str, defaults to 'price'
        User-defined name for the price column.
    quantity_col: str, defaults to 'quantity'
        User-defined name for the quantity column.
    date_col: str, defaults to 'month'
        User-defined name for the date column.
    product_id_col: str, defaults to 'id'
        User-defined name for the product ID column.
    characteristics: list of str, defaults to None
        The names of the characteristics columns.
    groups: list of str, defaults to None
        The names of the groups columns.
    method: str, defaults to 'all'
        Options: {'all', 'geks', gk', 'tpd', 'tdh'}

        The multilateral method to apply. The 'all' option uses the
        GEKS paired with a bilateral, GK and TPD index.
    bilateral_method: str, defaults to 'tornqvist'
        Options: {'carli', 'jevons', 'dutot', 'laspeyres', 'paasche',
        'geom_laspeyres', 'geom_paasche', 'drobish', 'marshall_edgeworth',
        'palgrave', 'fisher', 'tornqvist', 'walsh', 'sato_vartia', 'lowe',
        'geary_khamis_b', 'rothwell'}

        The bilateral method to pair with `method='geks'`.
    td_engine: str, defaults to 'numpy'
        Options: {'numpy', 'statsmodels', 'sklearn', 'pyspark'}

        Engine to use for wls computation with `method='tpd'`.
    reference_month: int or str, defaults to 1
        The month to use as the reference month for the multilateral methods. An
        integer specifies the position while a string species the month in the
        format 'YYYY-MM'.
    plot: bool, defaults to False
        Boolean parameter on whether to plot the resulting timeseries for price
        indices.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the timeseries and index values.
    """
    method, bilateral_method = method.lower(), bilateral_method.lower()

    valid_methods =  {'all', 'geks', 'gk', 'tpd', 'tdh'}
    valid_bilateral_methods = {
        'carli', 'jevons', 'dutot', 'laspeyres', 'lowe',
        'paasche', 'geom_laspeyres', 'geom_paasche', 'drobish',
        'marshall_edgeworth', 'palgrave', 'fisher', 'tornqvist',
        'walsh', 'sato_vartia', 'geary_khamis_b', 'tpd', 'rothwell'
    }

    if method not in valid_methods:
        raise ValueError("Invalid option, please select a valid method.")

    if method in {'all', 'geks'} and bilateral_method not in valid_bilateral_methods:
        raise ValueError("Invalid option, please select a valid bilateral method for GEKS.")

    args = (price_col, quantity_col, date_col, product_id_col)

    if groups:
        return (
            df
            .groupby(groups)
            .apply(
                    lambda df_group: multilateral_methods(
                        df_group,
                        *args,
                        characteristics=characteristics,
                        method=method,
                        bilateral_method=bilateral_method,
                        td_engine=td_engine,
                        reference_month=reference_month,
                        plot=plot
                    )
            )
            .reset_index()
            .rename({'level_1': 'month'}, axis=1)
        )
    if quantity_col not in df.columns:
        df[quantity_col] = 1
    if bilateral_method not in ('jevons', 'carli', 'dutot'):
        # Calculate weights for each item in each period.
        df = _weights_calc(df, *args)

    # Obtain unique time periods present in the data.
    periods = df[date_col].unique()

    if isinstance(reference_month, str):
        reference_month = periods.index(reference_month) + 1

    if method == 'all':
        index_vals = {
            f'index_value_geks': geks(df, *args, bilateral_method),
            'index_value_gk': geary_khamis(df, *args),
            'index_value_td': time_dummy(df, *args, characteristics, engine=td_engine)
        }
    elif method == 'geks':
        index_vals = geks(df, *args, bilateral_method)
    elif method == 'gk':
        index_vals = geary_khamis(df, *args)
    elif method == 'tpd':
        index_vals = time_dummy(df, *args, None, engine=td_engine)
    elif method == 'tdh':
        if not characteristics:
            raise ValueError("Characteristics required for TDH.")
        else:
            index_vals = time_dummy(df, *args, characteristics, engine=td_engine)
    output_df = (
        pd.DataFrame(
            index_vals,
            index=periods
        )
        .rename({0: 'index_value'}, axis=1)
    )
    output_df.sort_index(inplace=True)
    if refence_month != 1:
        output_df = output_df / output_df.iloc[refence_month - 1]
    if plot:
        sns.set(rc={'figure.figsize':(11, 4)})
        (output_df * 100).plot(linewidth=2)
    return output_df



from gettext import find
from setuptools import setup, find_packages

setup(
    name='PriceIndexCalc',
    version='0.1-dev5',
    description='Price Index Calculator using bilateral and multilateral methods',
    author='Dr. Usman Kayani',
    url='https://github.com/drrobotk/PriceIndexCalc',
    author_email='usman.kayani@ons.gov.uk',
    license='MIT',
    packages=find_packages(where="src"),
    package_dir={'': 'src'},
    install_requires=['pandas', 'numpy', 'scipy'],
    include_package_data=True,
)


"""
Provides the following multilateral methods:

* :func:`time_dummy`
* :func:`geary_khamis`
* :func:`geks`
paired with
    * :func:`carli`
    * :func:`jevons`
    * :func:`dutot`
    * :func:`laspeyres`
    * :func:`paasche`
    * :func:`geom_laspeyres`
    * :func:`geom_paasche`
    * :func:`drobish`
    * :func:`marshall_edgeworth`
    * :func:`palgrave`
    * :func:`fisher`
    * :func:`tornqvist`
    * :func:`walsh`
    * :func:`sato_vartia`
    * :func:`geary_khamis_b`
    * :func:`tpd`
    * :func:`rothwell`

The TDH/TPD methods are model-based multilateral index number methods
which have been proposed to incorporate scanner data. They are part of
many multilateral methods motivated by an attempt to minimize the risk
of chain drift, particularly within a window, while maximizing the
number of matches in the data.

TDH index is used when information on item characteristics are
available, and the TPD index when this information is lacking. The
TDH produces an explicit hedonic price index, while the TPD produces
an implicit hedonic price index, which are both estimated on the
pooled data of one or more periods via an application of expenditure
shares weighted least squares regression.
"""

from typing import List, Sequence, Optional
from itertools import combinations

import pandas as pd
import numpy as np
from scipy.stats.mstats import gmean

from .bilateral import *
from .helpers import diag, _weights_calc
from .weighted_least_squares import wls

__author__ = ['Dr. Usman Kayani']

def geks(
    df: pd.DataFrame,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str='month',
    product_id_col: str='id',
    bilateral_method: str = 'tornqvist',
) -> List:
    """
    Obtain the GEKS indices paired with a bilateral method for a given dataframe.

    Calculate the index values using a for loop to determine the matrix of
    bilaterals, where we exploit the symmetry condition a_{i j} = 1/a_{j i} and
    a_{i i} = 1 to save computation time, followed by a geometric mean.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing the data.
    price_col : str, optional
        Name of the column containing the price information.
    quantity_col : str, optional
        Name of the column containing the quantity information.
    date_col : str, optional
        Name of the column containing the date information.
    product_id_col : str, optional
        Name of the column containing the product id information.
    bilateral_method : str, optional

    Returns
    -------
    List
        List of the GEKS indices.
    """
    # Reverse unstack from dynamic window func.
    #df = df.stack().reset_index([date_col, product_id_col])

    # Get unique periods and length of time series.
    periods = df[date_col].unique()
    no_of_periods = len(periods)

    if bilateral_method != 'tpd':
        # Obtain bilateral function for bilateral method.
        bilateral_func = globals()[bilateral_method]

    # Intialize matrix for bilateral pairs.
    pindices = np.zeros((no_of_periods, no_of_periods))

    for month_idx in combinations(range(no_of_periods), 2):
        # Get period index for base and current month, and slice df for these
        # months.
        i, j = month_idx
        df_base = df.loc[df[date_col] == periods[i]]
        df_curr = df.loc[df[date_col] == periods[j]]

        # Make sure the sample is matched for given periods.
        df_base = df_base[df_base[product_id_col].isin(df_curr[product_id_col])]
        df_curr = df_curr[df_curr[product_id_col].isin(df_base[product_id_col])]

        if bilateral_method == 'tpd':
            # Use multilateral TPD method with two periods.
            df_matched = (
                pd.concat([df_base, df_curr])
                .drop_duplicates()
                .drop(columns='weights') 
            )
            # Recalculate weights for matched df.
            df_matched = _weights_calc(df_matched)
            # Append values to upper triangular of matrix.
            pindices[i, j] = time_dummy(df_matched)[-1]
        else:
            # Find price and quantity vectors of base period and current period.
            p_base = df_base[price_col].to_numpy()
            p_curr = df_curr[price_col].to_numpy()
            data = (p_base, p_curr)

            # Get quantities for bilateral methods that use this information.
            if bilateral_method in {
                'laspeyres', 'drobish', 'marshall_edgeworth',
                'geom_laspeyres', 'tornqvist', 'fisher',
                'walsh', 'sato_vartia', 'geary_khamis_b', 
                'rothwell', 'lowe'
            }:
                q_base = df_base[quantity_col].to_numpy()
                data += (q_base, )
            if bilateral_method in {
                'paasche', 'drobish','palgrave',
                'marshall_edgeworth', 'geom_paasche', 'tornqvist',
                'fisher', 'walsh', 'sato_vartia',
                'geary_khamis_b'
            }:
                q_curr = df_curr[quantity_col].to_numpy()
                data += (q_curr, )

            # Determine the bilaterals for each base and current period and
            # append to upper tringular of matrix.
            pindices[i, j] = bilateral_func(*data)

    # Exploit symmetry conditions for matrix of bilaterals.
    pindices_sym = np.copy(pindices.T)
    mask = pindices_sym != 0
    pindices_sym[mask] = 1/pindices_sym[mask]
    pindices += pindices_sym + np.identity(no_of_periods)

    # Calculate geometric mean for the unnormalized price levels.
    pgeo = gmean(pindices)

    # Normalize to first period.
    return pd.Series(
        pgeo/pgeo[0],
        index=periods,
    )


def time_dummy(
    df: pd.DataFrame,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str = 'month',
    product_id_col: str = 'id',
    engine: str = 'numpy'
) -> List:
    """Obtain the time dummy indices for a given dataframe.

    Calculates the time dummy indices using a formula with weighted least
    squares regression.  When passed with characteristics, this function returns
    the Time Dummy Hedonic indices. When passed without it returns the Time
    Product Dummy indices.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing the data.
    price_col : str, optional
        Name of the column containing the price information.
    quantity_col : str, optional
        Name of the column containing the quantity information.
    date_col : str, optional
        Name of the column containing the date information.
    product_id_col : str, optional
        Name of the column containing the product id information.
    engine : str, optional
        Name of the engine to use for the calculation.

    Returns
    -------
    List
        List of the time dummy indices.
    """
    # Reverse unstack from dynamic window func.
    df = df.stack().reset_index([date_col, product_id_col])

    # Set the dtype for ID columns, in case it is numerical.
    df[product_id_col] = df[product_id_col].astype(str)

    # Calculate logarithm of the prices for each item for dependent variable.
    df['log_price'] = np.log(df[price_col])

    # Get time series for output index.
    time_series = df[date_col].unique()

    # Get terms for wls regression where characteristics are used if available.
    non_time_vars = [product_id_col]

    model_params = wls(
        df,
        dependent_var='log_price',
        independent_vars=[date_col, *non_time_vars],
        engine=engine
    )

    # Get indices from the time dummy coefficients & set first = 1.
    is_time_dummy = model_params.index.str.contains(date_col)

    return pd.Series(
        [1, *np.exp(model_params.loc[is_time_dummy])],
        index=time_series,
    )


def geary_khamis(
    df: pd.DataFrame,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str = 'month',
    product_id_col: str = 'id',
    method_type: str = 'matrix',
) -> List:
    r"""Obtain the Geary-Khamis indices for a given dataframe.

    Calculates the Geary-Khamis indices using matrix operations.
    
    Parameters
    ----------
    price_col : str, defaults to 'price'
        User-defined price column name.
    quantity_col : str, defaults to 'quantity'
        User-defined quantity column name.
    product_id_col : str, defaults to 'product_id'
        The column name containing product ID values or product names.
    method_type: str, defaults to 'matrix'
        Options: {'matrix', 'iterative'}

        The method type to use for the GK computation.
    Returns
    -------
    List
        The sorted list of indices for each group.

    Notes
    -----
    For Geary-Khamis with the matrix method, we can determine the
    quality adjustment factors by solving the system of equations:

    .. math::

            \vec{b}=\left[I_{N}-C+R\right]^{-1} \vec{c}

    where :math:`\vec{c} = [1,0,\ldots, 0]^T` is an :math:`N \times 1`
    vector and :math:`R` is an :math:`N \times N` matrix given by,

    .. math::

            R=\left[\begin{array}{cccc}
                1 & 1 & \ldots & 1 \\
                0 & \ldots & \ldots & 0 \\
                \vdots & & & \vdots \\
                0 & \ldots & \ldots & 0
            \end{array}\right]

    and :math:`C` is the :math:`N \times N` matrix defined by,

    .. math::

            C=\hat{q}^{-1} \sum_{t=1}^{T} s^{t} q^{t \mathbf{T}}

    where :math:`\hat{q}^{-1}` is the inverse of an :math:`N \times N`
    diagonal matrix :math:`\hat{q}`, where the diagonal elements are the
    total quantities purchased for each good over all time periods,
    :math:`s^{t}` is a vector of the expenditure shares for time period
    :math:`t`, and :math:`q^{t \mathbf{T}}` is the transpose of the
    vector of quantities purchased in time period :math:`t`.

    Once the :math:`\vec{b}` vector has been calculated, the price
    levels can be computed from the equation:

    .. math::

            P_{t} =\frac{p^{t} \cdot q^{t}}{ \vec{b} \cdot q^{t}}

    The price index values can be determined by normalizing the price
    levels by the first period as,

    .. math::

            I_{t} = \frac{P_{t}}{P_{0}}

    References
    ----------
    Diewart, W. E, and Kevin, F. (2017). Substitution Bias in
    Multilateral Methods for CPI Construction Using Scanner Data.
    Discussion Paper 1702. Department of Economics, University of
    British Columbia.
    """
    if method_type not in ('matrix', 'iterative'):
        raise ValueError('The method type must be `matrix` or `iterative`')

    # We need to deal with missing values and reshape the df for the
    # required vectors and matrices.
    df = _matrix_method_reshape(df)

    # Get number of unique products for the size of the vectors and
    # matrices.
    N = len(df.index.unique(level=product_id_col))

    # Matrices for the prices, quantities and weights.
    prices = df.loc[price_col]
    quantities = df.loc[quantity_col]
    weights = df.loc['weights']

    # Use iterative method directly if specified.
    if method_type == 'iterative':
        return _geary_khamis_iterative(prices, quantities)

    # Inverse of diagonal matrix with total quantities for each good over all
    # time periods as diagonal elements and matrix product of weights and
    # transpose of quantities to produce a square matrix, both required for C
    # matrix.
    q_matrix_inverse = np.diag(1/quantities.T.sum())
    prod_weights_qt_matrix = weights @ quantities.T

    # Product of above matrices to give the C square matrix, with the fixed
    # identity and R matrix, and c vector all required determine the quality
    # adjustment factors b.
    C_matrix = q_matrix_inverse @ prod_weights_qt_matrix
    R_matrix = np.zeros(shape=(N, N))
    R_matrix[:1] = 1
    
    # Define combo matrix used for isolating the singular matrices and
    # calculating the index values from the combo matrix `I_n - C + R`.
    combo_matrix = np.identity(N) - C_matrix + R_matrix
    
    if abs(np.linalg.det(combo_matrix)) <= 1e-7:
        # Fallback to iterative method for singular matrices.
        return _geary_khamis_iterative(prices, quantities)
    else:
        # Primary matrix method for non-singular matrices.
        return _geary_khamis_matrix(prices, quantities, combo_matrix)


def _matrix_method_reshape(df: pd.DataFrame) -> pd.DataFrame:
    """
    Reshape df for matrix method and deal with missing values.
    We first drop columns which contain all missing values, transpose
    the dataframe and then fill the remaining missing values with zero,
    to deal with missing items in some periods.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to reshape.

    Returns
    -------
    pd.DataFrame
        The reshaped dataframe.
    """
    return df.dropna(how='all', axis=1).T.fillna(0)


def _geary_khamis_iterative(
    prices: pd.DataFrame,
    quantities: pd.DataFrame,
    no_of_iterations: int = 100,
    precision: float = 1e-8,
) -> pd.Series:
    """
    Geary-Khamis iterative method.
    
    Parameters
    ----------
    prices : pd.DataFrame
        The price dataframe.
    quantities : pd.DataFrame
        The quantity dataframe.
    no_of_iterations : int, defaults to 100
        The number of iterations to perform.
    precision : float, defaults to 1e-8
        The precision to use for the iterative method.

    Returns
    -------
    pd.Series
        The price index values.
    """
    # Initialise index vals as 1's to find the solution with iteration.
    price_levels = pd.Series(1.0, index=prices.columns)
    quantity_share = quantities.T / quantities.sum(axis=1)

    # Iterate until we reach the set level of precision, or after a set
    # number of iterations if they do not converge.
    for _ in range(no_of_iterations):
        # Obtain matrices for iterative calculation.
        deflated_prices = prices / price_levels
        factors = diag(deflated_prices @ quantity_share)

        # Calculate new price levels from previous value.
        new_price_levels = (
            diag(prices.T @ quantities)
            .div(quantities.T @ factors)
            .squeeze()
        )

        pl_abs_diff = abs(price_levels - new_price_levels)
        
        if (pl_abs_diff <= precision).all():
            # Break loop when we reach given precision for final price levels.
            break
        else:
            # Otherwise set price level for next iteration.
            price_levels = new_price_levels

    # Normalize by first period for final output.
    return price_levels / price_levels.iloc[0]


def _geary_khamis_matrix(
    prices: pd.DataFrame,
    quantities: pd.DataFrame,
    combo_matrix: pd.DataFrame,
) -> pd.Series:
    """
    Geary-Khamis matrix method.

    Parameters
    ----------
    prices : pd.DataFrame
        The price dataframe.
    quantities : pd.DataFrame
        The quantity dataframe.
    combo_matrix : pd.DataFrame
        The combo matrix.

    Returns
    -------
    pd.Series
        The price index values.
    """
    # Calculation of the vector b (factors) required to produce the
    # price levels. Corresponds to `b = [I_n - C + R]^-1 [1,0,..,0]^T`.
    # We use the Moore-Penrose inverse for the matrix inverse.
    factors = np.linalg.pinv(combo_matrix) @ np.eye(len(prices.index), 1)

    # Determine price levels to compute the final index values.
    price_levels = diag(prices.T @ quantities).div(quantities.T @ factors)

    # Normalize price levels to first period for final index values.
    index_vals = price_levels / price_levels.iloc[0]

    # Output as Pandas series for dynamic window.
    return index_vals.iloc[:, 0]
